/* maze.c   a maze generator
   Jon Bennett
        jcrb@cs.cmu.edu
   */

/* the maze is generated by making a list of all the internal hedges
   and randomizing it, then going lineraly through the list, we take a
   hedge and se if the maze squares adjacent to it are already connected
   (with find) is not the we connect them (with link), this prevents us
   from creating a maze with a cycle because we will not link two maze
   squares that are already connect by some path */
#include <stdio.h>

#define DOWN 1
#define RIGHT 2

struct maze_loc{
    int rank;
    int x,y;
    struct maze_loc *ptr;
};
struct hedge_loc{
    int x,y,pos,on;
};

struct maze_loc *maze;
struct hedge_loc *hedge;
struct hedge_loc **hedge_list;

void link(a,b)
 struct maze_loc *a,*b;
{
    if(a->rank == b->rank){
        a->ptr=b;
        b->rank++;
        return;
    }
    if(a->rank > b->rank){
        b->ptr=a;
        return;
    }
    a->ptr=b;
}
struct maze_loc *find(a)
 struct maze_loc *a;
{
    if(a != a->ptr){
        a->ptr = find(a->ptr);
    }
    return a->ptr;
}

main(argc,argv)
 int argc;
 char **argv;
{
    int x,y,i,j,k,l,tmp;
    struct maze_loc *a,*b;
    struct hedge_loc *htmp;
    int cycles;
    int c;

    if(argc!=4) {
        printf("use: genmaze width height number_cycles\n");
        exit(1);
    }

    srandom(time(0));
    x=atoi(argv[1]);
    y=atoi(argv[2]);
    cycles=atoi(argv[3]);
    

    /*malloc the maze and hedges */
    maze=(struct maze_loc *)malloc(sizeof(struct maze_loc)*x*y);
    hedge=(struct hedge_loc *)malloc(sizeof(struct
hedge_loc)*((x*(y-1))+((x-1)*y)));
    hedge_list=(struct hedge_loc **)malloc(sizeof(struct hedge_loc
*)*((x*(y-1))+((x-1)*y)));

    /*init maze*/

    for(j=0;j<y;j++){
        for(i=0;i<x;i++){
            maze[x*j+i].x = i;
            maze[x*j+i].y = j;
            maze[x*j+i].ptr = &maze[x*j+i];
            maze[x*j+i].rank=0;
        }
    }

    /*init hedges*/
    for(j=0;j<(y-1);j++){
        for(i=0;i<x;i++){
            hedge[x*j+i].x = i;
            hedge[x*j+i].y = j;
            hedge[x*j+i].pos=DOWN;
            hedge[x*j+i].on=1;
            hedge_list[x*j+i]= &hedge[x*j+i];
        }
    }
    k=x*(y-1);

    for(j=0;j<y;j++){
        for(i=0;i<(x-1);i++){
            hedge[k+(x-1)*j+i].x = i;
            hedge[k+(x-1)*j+i].y = j;
            hedge[k+(x-1)*j+i].pos=RIGHT;
            hedge[k+(x-1)*j+i].on=1;
            hedge_list[k+(x-1)*j+i]= &hedge[k+(x-1)*j+i];
        }
    }

    k=(x*(y-1))+((x-1)*y);

    /*randomize hedges*/
    for(i=(k-1);i>0;i--){
        htmp=hedge_list[i];
        j=random()%i;
        hedge_list[i]=hedge_list[j];
        hedge_list[j]=htmp;
    }
    fflush(stdout);

    l=k;

    /*create maze*/
    for(i=0;i<l;i++){
        j=hedge_list[i]->x;
        k=hedge_list[i]->y;
        a=find(&maze[x*k+j]);
        if(hedge_list[i]->pos==DOWN){
            b=find(&maze[x*(k+1)+j]);
        } else {
            b=find(&maze[x*k+j+1]);
        }
        if(a!=b){
            link(a,b);
            hedge_list[i]->on=0;
        }
    }

    /* Randomly knock down cycles additional hedges */
    for(i=0, c=0;(i<l) && (c<cycles);i++){
        if (hedge_list[i]->on){
            hedge_list[i]->on=0;
            c++;
        }
    }

    printf("%i %i\n", x,y);

    printf("+");
    for(i=0;i<x;i++){
        printf("-+");
    }
    printf("\n");

    l=x*(y-1);

    for(j=0;j<y;j++){
        printf("|");
        for(i=0;i<(x-1);i++){
            if (j==0&&i==0){
                if(hedge[l+(x-1)*j+i].on){
                    printf("*|");
                } else {
                    printf("* ");
                }
            }
            else{
                if(hedge[l+(x-1)*j+i].on){
                    printf(" |");
                } else {
                    printf("  ");
                }
            }
        }
        
        if (j<(y-1))
            printf(" |\n|");
        else
            printf("X|\n+");

        if(j<(y-1)){
            for(i=0;i<x;i++){
               if(hedge[j*x+i].on){
                   printf("-+");
               } else {
                   printf(" +");
               }
            }
            printf("\n");
        }

    }

    for(i=0;i<x;i++){
        printf("-+");
    }
    printf("\n");
    
    return 0;

}
